<?php
/*
* @Prometeus Technology, a subsidiary of Pegasus Tech Supply
* @copyright Copyright (C) 2008-2013 Leonardo Tadei. All rights reserved.
* @license http://www.gnu.org/copyleft/gpl.html GNU/GPL, see gpl-3.0.txt
* ORM is free software. This version may have been modified pursuant
* to the GNU General Public License, and as distributed it includes or
* is derivative of works licensed under the GNU General Public License or
* other free or open source software licenses.
*
* ORM implementa persistencia no invasiva y por alcance en PHP.
*
*/

require_once("Config.php");

define('PersistentMaxObjectPerPage', '9999999999994');

class Persistent {

    // Conexión con PHP Data Objects
    private $pdo;

    // Archivo de configuración
    private $configxml;
    private $xmldoc;

    // Orden de las Colecciones
    private $collectionOrder = '';	// espera un atributo
    private $collectionAsc = '';	// ASC o DESC
    // Filtro de las Colecciones
    private $filterAttrib = array('');
    private $filterValue = array('');
    private $filterOper = array('=');
    private $filterExac = TRUE;

    // Tablas en la DB para validar
    private $tablas;

    // Flag para hidratar Objetos anidados o no (recontra útil para levantar id/nombre)
    public $recursiveObject = true;

    // Flag para mostrar info de depurado
    public $debug = false;

    // Versión
    private $version = 0.46;

    // Constructor. Determina el archivo de configuración y conecta a la DB.
    function __construct($configxml="Persistent.xml") {
        if( ! file_exists($configxml) ) { throw new Exception('No se encuentra el archivo de mapeo: '.$configxml); }
        $this->configxml = $configxml;
        // Carga al documento XML de mapeo acá, para hacerle LOAD solo una vez
        $this->xmldoc = new DOMDocument;
        // We don't want to bother with white spaces
        $this->xmldoc->preserveWhiteSpace = false;
        $this->xmldoc->Load($this->configxml);
// Debug if (defined(PDO_DB)) { print('Definida'.PDO_DB ); } else { print('inDefinida'); }
//var_dump($_SERVER);

        $this->pdo = new PDO(PDO_DB,DB_USER,DB_PASS);
        // Hace que los errores generen exepciones
		  $this->pdo->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION);

        // Guarda las tablas en la DB para validar la concordancia
        $sql = 'SHOW TABLES FROM `'.DB_NAME.'`';
        $stmt = $this->pdo->prepare($sql);
        $stmt->execute();
        while( $row = $stmt->fetch() ) {
            $this->tablas[] = $row[0];
        }
// Debug
/*			print("Debug tablas $sql: <pre>");
                print_r($this->tablas);
                print('</pre>');*/
    }

    // Guarda un Objeto
    function save($instancia) {
            $clase = get_class($instancia);
            $nulo = True;
            // Si no hay mapeo, entonces es el default id
            $oid = $this->classMapper($clase, 'ObjectId');
            // Si no hay mapeo, entonces es el default id
            $pk = $this->classMapper($clase, 'Table/PrimaryKey');

//Debug 	var_dump($instancia);
// Debug print("<br><b>Reflection</b> sobre: $clase | $instancia<br>");
            if($this->debug) { print ("Debug save: $clase | PK: $pk | OID: $oid<br>");}
            // Inicializa una posible relación 1-N
            $col1n = array();
            // Vector con los atributos
            $att = $this->getAttibutes($clase);
            foreach($att as $name) {
                $getter = $this->classMapper($clase, "Attributes/$name/getter");
                $campo = $this->classMapper($clase, "Attributes/$name/field");

                // Si un atributo es un objeto mapeado, debe llamar a save()
                // ...
                $type = '';
                $relation  = $this->classMapper($clase, "Attributes/$name/relation", $type);
                $att_name = '';
                $reference  = $this->classMapper($clase, "Attributes/$name/reference", $att_name);
                if(($this->debug)&&($relation !='')) { print ("Debug save $clase: Relation = $relation (type: $type)| Reference = $reference | Att_Name: $att_name<br>");}
                switch($relation) {
                    // Relación 1-1 (persiste un objeto anidado)
                    case "1-1":
                        // valida que haya una referencia mapeada
                        if ($reference != '' ) {
                                // Si es "composition", persiste al Objeto anidado
                                if($type == 'composition') {
                                        $oidtmp = $this->save($instancia->{$getter}() );
                                        // Carga en $a[$name] el OID del Objeto anidado para guardar la referencia
                                        $a[$campo] = $oidtmp;
                                        // Para que el Objeto anidado no parezca NULL
                                        $nulo = False;
                                } else {
                                        // si no, el default es "agregation", y solo guarda el OID del Objeto relacionado
                                        $clase_anidada = get_class($instancia->{$getter}());
                                        $oid_anidado = $this->classMapper($clase_anidada, 'ObjectId');
                                        $getter_anidado = $this->classMapper($clase_anidada, "Attributes/$oid_anidado/getter");
                                        // Obtiene el OID del Objeto anidado
                                        //$objtmp = $this->retrieve($clase_anidada, $instancia->{$getter}()->{$getter_anidado}());
                                        //$oidtmp = $objtmp->{$getter_anidado}();
                                        if($instancia->{$getter}() != NULL) {
                                                $oidtmp = $instancia->{$getter}()->{$getter_anidado}();
                                        } else {
                                                $oidtmp = NULL;
                                        }
                                        // Carga en $a[$name] el OID del Objeto anidado para guardar la referencia
                                        $a[$campo] = $oidtmp;
                                        // Para que el Objeto anidado no parezca NULL
                                        $nulo = False;
                                } // 	if($type == 'composition')
                                if($this->debug) { print ("Debug save: SubObjeto mapeado: $clase_anidada | OID: $oid_anidado | campo: $campo | getter: $getter_anidado | Valor OID: $oidtmp<br>"); }
//print("<br>Salida de save() = $oidtmp<br>");
                        } else { //if ($reference != '' )
                                throw new Exception( "Error: no hay Objeto de referencia mapedo para la relacion." );
                        }
                        break;
                    // Relación 1-N (persiste colecciones de objetos anidados)
                    case '1-N':
                        // valida que haya una referencia mapeada
                        if (($reference != '')&&($att_name != '')) {
                                // Lo guarda temporalmente, porque el objeto "1" puede no tener ID todavía, y hay que hacer esto después
                                $col1n[$name] = $instancia->{$getter}();
//print("(die)Save en 1-N: $clase Relation = $relation (type: $type)| Reference = $reference | Att_Name: $att_name<br>");die();
                        } else { //if ($reference != '' )
                                throw new Exception( "Error: no hay Objeto de referencia mapedo para la relacion." );
                        }
                        break;
                    default:
                        // Si es un atributo plano, lo agrega al vector
                        //$a[$name] = $instancia->{$getter}();
                        $a[$campo] = $instancia->{$getter}();
                }

                // Compara con el ObjectId para ver si es nulo
                if ($name != $oid) {
//print("Debug $name != $oid <br>");
                        //if (($a[$name] != NULL)) { $nulo = False; }
                        if(isset($a[$name]) && ($a[$name] != NULL)) { $nulo = False; }
                }

            } // foreach($att as $name)

// Debug atributos recibidos
//print("<pre>Atributos en save()\n");print_r($a);print("</pre>");

/*
 * Esto resultó ser una mala idea, porque si es una composición y los valores del objeto
 * compuesto pueden ser nulos, entonces lo borra y no respeta el ciclo de vida del
 * objeto padre.
 * El $type es el que indicaría esto, pero se usa el del propio Objeto hijo, que no tiene
 * relación alguna, porque el que la tiene es el mapeo del padre.
            // Si todos los atributos son nulos y tiene ID, entonces lo borro del repositorio
            // (esto no funciona si la lógica del Objeto impide asignar nulos...)
            // Si es una Composición, el ciclo de vida del Objeto compuesto tiene que respetear al del Objeto Padre.
var_dump('orm',$clase,$type);
            if (($type != 'composition') && ($nulo) && ($a[$oid]!=NULL)) {
                    //return $this->delete($clase, $a);
                if($this->debug) { print ("Debug save: llamada a delete() por datos NULL | Nulo: $nulo | OID: $oid<br>");  var_dump($a); print("<br>"); var_dump($nulo); }
                $resultado = $this->delete($instancia);
            }
 *
 */

            // Si el ID es nulo, es nuevo e inserta, sino actualiza.
            if ($a[$oid] == NULL) {
                    //print("Debug: entra insert. OID: $oid | PK: $pk - {$a[$oid]}<br>");
                    $resultado = $this->insert($clase, $a);
                    // Ahora que está persistido, puedo asignarle el OID
                    $setterOID = $this->classMapper($clase, "Attributes/$oid/setter");
                    $instancia->{$setterOID}($resultado);
            } else {
                    //print("Debug: entra update. OID: $oid | PK: $pk - {$a[$oid]}<br>");
                    $resultado = $this->update($clase, $a);
            }

            // Si la relación es 1-N, tengo que guardar primero el objeto "1" para
            // que tenga un OID que asignarle al objeto "N"
            foreach ($col1n as $k => $v) {
                    $type = '';
                    $relation  = $this->classMapper($clase, "Attributes/$k/relation", $type);
                    // Obtiene el getter del OID de la instancia 1
                    $getterID = $this->classMapper($clase, "Attributes/$oid/getter");
                    $oid2 = $this->classMapper($reference, 'ObjectId');
                    foreach ($v as $obj) {
                            // Obtiene el OID de la instancia 1
                            // Debug print("Debug: getterID $getterID");
                            $valID = $instancia->{$getterID}();
                            if($valID == '') { $valID = $resultado; }
                            // setea el valor de la instancia 1 en el atributo de la N
                            $att_name = '';
                            $reference  = $this->classMapper($clase, "Attributes/$k/reference", $att_name);
                            $setterN = $this->classMapper($reference, "Attributes/$att_name/setter");
                            if($type == 'composition') {
                                    // $obj es la instancia del objeto N en la colección del objeto 1
                                    $obj->{$setterN}($valID);
                                    // Guarda cada uno de los N
                                    $this->save($obj);
                            } else {
// Esto es horriblemente lento...
// se podría reemplazar por una llamada a un método que haga
// un UPDATE N SET rel=1 WHERE id = obj2->getID teniendo en cuenta que el ID
// del objetoN no cambia y puedo usar los contenidos en el objeto 1.
                                    // Si es 'agregation', debo garantizar que solo se actualice la referencia
                                    // Rehidrata los objetos del repositorio y solo le cambia la referencia
                                    $getter1 = $this->classMapper($reference, "Attributes/$oid2/getter");
                                    $obj2 = $this->retrieve($reference, $obj->{$getter1}() );
                                    // le pone la referencia al objeto 1 correspondiente
                                    $obj2->{$setterN}($valID);
                                    // Guarda cada uno de los N
                                    $this->save($obj2);
                            }
                            // Debug var_dump($obj);
                            if($this->debug) { print("Save Rel 1-N: $k ($type) con Obj1 = $valID | Ref: $reference / AttName: $att_name | setter: $setterN<br> "); }
                    }
            } // foreach ($col1n as $k => $v)

            // Devuelve el OID
            return $resultado;
    } // function save($instancia)

    // Borra un Objeto del repositorio
    function delete(&$obj) {
            if (is_object($obj)) {
                    $clase = get_class($obj);
            } else {
                    throw new Exception( "Error: $class debe ser un Objeto." );
            }
            // A partir de la clase, obtiene la tabla mapeada.
            $tabla = $this->classMapper($clase, 'Table/Name');
            // Valida que la tabla hallada exista o sale silenciosamente
            if( ! in_array($tabla, $this->tablas) ) { return True; }
            // A partir de la clase, obtiene el campo clave mapeado
            $pkey  = $this->classMapper($clase, 'Table/PrimaryKey');
            // Si el campo tiene la forma T.campo, se queda solo con el nombre del campo
            if(strpos($pkey, '.') !== false ) {
                    $pkey = substr($pkey, strpos($pkey, '.')+1, strlen($pkey));
            }
            // A partir de la clase, obtiene el OID mapeado
            $oid = $this->classMapper($clase, 'ObjectId');
            // Con el OID, obtiene el getter del atributo...
            $getter = $this->classMapper($clase, "Attributes/$oid/getter");
// Debug print("<br>Delete: Tabla = $tabla | PK = $pkey | OID = $oid | getter = $getter <br>");
            // ... y obtiene el campo clave a borrar
            $val = $obj->$getter();

            // Si hay un Objeto Relacionado y es una Composición, debe borrarlo también
            // ...
            // ...
            // Vector con los atributos
            $att = $this->getAttibutes($clase);
            foreach($att as $name) {
                    $tipo = NULL;
                    $relation  = $this->classMapper($clase, "Attributes/$name/relation", $tipo);
                    if($tipo == "composition") {
                            switch($relation) {
                                    case '1-N':
                                            // La copmposición 1-N tiene que borrar todos los N
                                            $getterN = $this->classMapper($clase, "Attributes/$name/getter");
                                            foreach($obj->$getterN() as $objN) {
                                                    if($this->debug) { print("Delete ObjN: {$objN->getId()}, {$objN->getNom()}<br>");	}
                                                    $this->delete($objN);
                                            }
//print("(die)Delete: Tipo: $tipo | Rel: $relation | Name: $name | getter: $getterN <br>"); die();
                                            break;
                                    default: // Para relaciones 1-1
                                            // Obtiene el getter del atributo que es el Objeto relacionado
                                            $relget = $this->classMapper($clase, "Attributes/$name/getter");
                                            // Obtiene el OID del atributo relacionado y lo borra
/* var_dump('ORM',$obj->$relget());
                                            $this->delete( $obj->$relget() ); */

                                            // Se usa una variable para invocar a delete porque devuelve un warning strict, de que solo variables pueden ser pasadas por referencia
                                            $trashobject = $obj->$relget();
// var_dump('ORM',$trashobject);
                                            $this->delete( $trashobject );
                                            // Debug print("<br><b>Relacion</b>: $tipo | $relget | {$obj->$relget()} <br>");
                            } // switch
                    } // if($tipo == "composition")
            } // foreach

            $sql = "DELETE FROM `$tabla` WHERE $pkey='$val'";

            // Si el Objeto no existe en el repositorio, sale
            if($val == NULL) {
                    throw new Exception( "Error: Objeto inexistente en el Repositorio." );
            }

            if($this->debug) { print("Debug delete: $sql <br>");  /* $rompe = 3/0; */ }

// Debug ! Sale para no borrar de veras mientras desarrollo
//exit();

            // Si borrar no afecta al menos una fila, entonces falla.
            if( $this->pdo->exec($sql) === 0 ) {
                    $error = $this->pdo->errorInfo();
                    throw new Exception( "Error: class $clase | {$error[0]} - {$error[1]} : {$error[2]}." );
            }
            // Luego de borrado, lo libera de la RAM
            $obj = NULL;

            return True;
    } // function delete()

    // Hidrata un Objeto del repositorio
    function retrieve($classname, $id) {
        // Si $id es NULL, no hay ninguna instancia que devolver
        if($id == NULL) { return NULL; }

        // Crea una nueva instancia del Objeto a hidratar (así está disponible para 1-N)
        $newobj = new $classname;

        // A partir de la clase, obtiene la tabla mapeada.
        $tabla = $this->classMapper($classname, 'Table/Name');
        // Obtiene el SELECT correspondiente a esta clase
        $sql = $this->select($classname, $id);

        // Valida que la tabla hallada exista o tenga mapeado un select y sale con excepción
        //if( ! in_array($tabla, $this->tablas) ) {
        if(($sql == '') && ( ! in_array($tabla, $this->tablas) ) ) {
             throw new Exception( "Error: Mapeo inexistente en el repositorio. Clase: $classname");
        }

        //$sql = "$sql $pkey=$id ";
        $stmt = $this->pdo->query($sql);

        if($this->debug) {
            print ("<br>Debug retrieve (1): $sql<br>");
        }

        if ($stmt == NULL) {
            $error = $this->pdo->errorInfo();
            throw new Exception( "Error: $sql | {$error[0]} - {$error[1]} : {$error[2]}.");
        }

        // Objeto STD conteniendo el resultado de la query
        $obj = $stmt->fetch(PDO::FETCH_OBJ);

        if ($obj === false) {
            // Cambio 24/09/2012 . En vez de fallar si no se encuentra, devuelve NULL
            // throw new Exception( "Error: No existe una instancia de $classname con OID $id en el repositorio.");
            return NULL;
        } else {
            // Los mapeos están ahora en un método privado
            $newobj = $this->hydrate($newobj, $obj);
            return $newobj;
        }
    } // function retrieve

    // Muestra la estructura de la DB asociada a la Clase del Objeto
    function struct($class) {
            $clase = get_class($class);
            // A partir de la clase, obtiene la tabla mapeada.
            $tabla = $this->classMapper($clase, 'Table/Name');

            $sql = "SHOW FIELDS FROM `$tabla`";
            if($this->debug) {
                    print("Debug struc: $sql <br>");
            }
            foreach ($this->pdo->query($sql) as $row) {
                    print("<pre>");print_r($row);	print("</pre>");
            }
    }

    // Genera una entrada XML con los default para la clase
    function defaultXML($class) {
            if (is_object($class)) {
                    $clase = get_class($class);
            } else {
                    $clase = $class;
            }

            $tabla = $this->classMapper($clase, 'Table/Name');


            $buffer = "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>
<Persistent>\n";
            $buffer .= "\t<$clase>\n";

            $buffer .= "\t\t<Table>\n";
            $buffer .= "\t\t\t<Name>$tabla</Name>\n";
            $buffer .= "\t\t\t<PrimaryKey>id</PrimaryKey>\n";
            $buffer .= "\t\t\t<Select>SELECT * FROM $tabla</Select>\n";
            $buffer .= "\t\t</Table>\n";

            $buffer .= "\t\t<ObjectId>id</ObjectId>\n";

            $buffer .= "\t\t<Attributes>\n";

            $att = $this->getAttibutes($class);
            foreach ($att as  $name) {
                    // Obtiene Atributos
                    $buffer .= "\t\t\t<$name>\n";
                    $buffer .= "\t\t\t\t<setter>set".ucfirst($name)."</setter>\n";
                    $buffer .= "\t\t\t\t<getter>get".ucfirst($name)."</getter>\n";
                    $buffer .= "\t\t\t\t<field>".$name."</field>\n";
                    $buffer .= "\t\t\t</$name>\n";
            }

            $buffer .= "\t\t</Attributes>\n";


            $buffer .= "\t</$clase>\n";
            $buffer .= "</Persistent>\n";

            $buffer .= "\n\n // Borrar todo lo que coincida con la Clase y la Tabla";

            return $buffer;
    }

    function getVersion() {
            return $this->version;
    }

    function __destruct() {
            $this->pdo = null;
    }

// --------------------------------------------------------------------------------
// --                                                                           --
// --                          Manejo de Colecciones                            --
// --                                                                           --
// --------------------------------------------------------------------------------
    // Asigna el orden de la colección
    function setCollectionOrder($attrib, $asc=true) {
            $this->collectionOrder = "$attrib";
            // Si no es Ascendete, explicita Descencente
            if($asc) { $this->collectionAsc = ''; }
            else { $this->collectionAsc = ' DESC '; }
    }
    // Devuelve al programador el orden de la Colección
    function getCollectionOrder() {
            return $this->collectionOrder.'|'.$this->collectionAsc;
    }

    // Devuelve si el orden es ascendente o descendente
    private function getMapperOrder($classname) {
            // Si no hay orden, devuelve vacío y sale
            if($this->collectionOrder == '') {
                    return '';
            } else {
                    // Determina el campo
                    if ($this->classMapper($classname, "Attributes/{$this->collectionOrder}/field") != '') {
                                    $campo = $this->classMapper($classname, "Attributes/{$this->collectionOrder}/field");
                            } else {
                                    $campo = $this->collectionOrder;
                            }
                    // Devuelve el filtro
                    //return " ORDER BY `$campo` {$this->collectionAsc} ";
                    return " ORDER BY $campo {$this->collectionAsc} ";
            }
                    // Si hay filtro, lo devuelve mapeado si corresponde
    }

    // Determina múltiples filtros para la Colección
    function addCollectionFilter($attrib, $val, $oper='=') {
        // Valida recibir un operador válido
        $operadoresValidos = array('=', '>', '>=', '<', '<=', '<>', 'LIKE', 'LIKE%', '%LIKE');
        if (! in_array($oper, $operadoresValidos)) {
            throw new Exception("Operador invalido: $oper");
        }
        // Agrega un filtro con el operador recibido en el mensaje
        if ($attrib != '') {
                $this->filterAttrib[] = $attrib;
                $this->filterValue[] = $val;
                $this->filterOper[] = strtoupper($oper);
        }
        return TRUE;
    }
    // Pone a cero los filtros
    function resetCollectionFilter() {
        $this->filterAttrib = array('');
        $this->filterValue = array('');
        $this->filterOper = array('=');
        return TRUE;
    }
    // Determina un filtro para la Colección (dejado por retrocompatibilidad)
    function setCollectionFilter($attrib, $val, $exac=true) {
        // Prepara un filtro para la coleeción por el atributo dado, con el valor dado.
        // Si es exacto hace =, y si no hace 'LIKE'
        if ($attrib != '') {
            /*$this->filterAttrib = $attrib;
            $this->filterValue = $val;
            $this->filterExac = $exac;*/
            $this->resetCollectionFilter();
            $oper = '=';
            if (! $exac) { $oper = 'LIKE'; }
            $this->addCollectionFilter($attrib, $val, $oper);
        } else {
            /*$this->filterAttrib = '';
            $this->filterValue = '';
            $this->filterExac = true;*/
            $this->resetCollectionFilter();
        }
        return True;
    }

    // Devuelve al programador el filtro de la colección activo
    function getCollectionFilter() {
        $sal = '';
        foreach ($this->filterAttrib as $k => $v) {
            $sal .= $v.'|'.$this->filterValue[$k].'|'.$this->filterOper[$k].'# ';
        }
        //return $this->filterAttrib .'|'.$this->filterValue.'|'.$this->filterOper;
        return $sal;
    }

    // Arma el filtro para la query según el mapeo
    private function getMapperFilter($classname, $query) {
        $filtro = '';
        // los filtros son ahora un vector que hay que recorrer
        // Asumo que por cada filterAttrib, están los corresponientes filterValue y filterOrder
        foreach ($this->filterAttrib as $k => $v) {
            if($v != '') {
                // Si hay filtro, lo devuelve mapeado si corresponde
                // Si recibe como mensaje opcional una query, se fija si corresponde WHERE o AND (es por
                // las querys mapeadas que pueden tener cualquier forma)
// Ojo!
// http://dev.mysql.com/doc/refman/5.0/es/problems-with-alias.html

                // Determina el operador
                $ope = $this->filterOper[$k];
                $comodin1 = '';
                $comodin2 = '';
                if($this->filterOper[$k] == 'LIKE') {
                    $comodin1 ='%';
                    $comodin2 ='%';
                }
                if($this->filterOper[$k] == '%LIKE') {
                    $ope = 'LIKE';
                    $comodin1 ='%';
                }
                if($this->filterOper[$k] == 'LIKE%') {
                    $ope = 'LIKE';
                    $comodin2 ='%';
                }
                // Determina el campo
                if ($this->classMapper($classname, "Attributes/{$this->filterAttrib[$k]}/field") != '') {
                        $campo = $this->classMapper($classname, "Attributes/{$this->filterAttrib[$k]}/field");
                    } else {
                        $campo = $this->filterAttrib[$k];
                    }
                // Analiza la query preexistente, si corresponde
                $where = ' WHERE ';
                // Si el select está mapeado o es la segunda iteración,
                // puede tener ya un WHERE... en ese caso hay que poner un AND en lugar de un WHERE
                $tmpquery = $query . $filtro;
                if( strrpos($tmpquery, 'WHERE') !== false ) {
                    $where = ' AND ';
                }
                //$filtro =  " $where `$campo` $ope '$comodin{$this->filterValue}$comodin' ";
                $filtro .=  " $where $campo $ope '$comodin1{$this->filterValue[$k]}$comodin2' ";
                // Devuelve el filtro
            }
        } // foreach
        return $filtro;
    }

    // Devuelve una Colección ya creada conteniendo los Objetos.
    // Depende de que la Colección tenga un método add() [implements IteratorAggregate].
    // Devuelve la cantidad de Objetos Totales en la Colección, sin tener en cuenta la paginación
    // pero sí los filtros
    function retrieveCollection($classname, &$collection, $page=1, $objPerPage=10) {
        // 9999999999994 es el máximo $objPerPage que funciona
        // Podría hacer que si son 0 ambos, la query no tenga LIMIT.. pero tener LIMIT es el secreto del éxito de PHPMyAdmin

        // Puede tener mapeado un SELECT distinto al default
        /*$select = $this->classMapper($classname, 'Table/Select');
        if($select == '') { $select = "SELECT * FROM `$tabla`"; }*/
        $select = $this->select($classname);

        // A partir de la clase, obtiene la tabla mapeada.
        $tabla = $this->classMapper($classname, 'Table/Name');
        // Valida que la tabla hallada exista o que tenga un SELECT mapeado y sale con excepción
        if(($select == '') && ( ! in_array($tabla, $this->tablas) ) ) {
                 throw new Exception( "Error: Mapeo inexistente en el repositorio. Clase $classname");
        }

        // Puede estar filtrada
        $filter = $this->getMapperFilter($classname, $select);
        // Si la Colección tiene orden, lo aplica
        $order = $this->getMapperOrder($classname);

        // Total de Items en la colección: si está filtrada, el total con filtro. si no está filtrada, el total
        // A partir de la clase, obtiene el campo clave
/*		$pkey  = $this->classMapper($clase, 'Table/PrimaryKey');
        $sql_tot = "SELECT COUNT($pkey) as objtot FROM `$tabla` $filter";
        $stmt = $this->pdo->prepare($sql_tot);
        $stmt->execute();
        $row = $stmt->fetch();
        $total = $row['objtot'];
        $stmt = NULL;*/

        $sql_tot = $this->classMapper($classname, 'Table/Count', $campo);
        $stmt = $this->pdo->prepare("$sql_tot $filter");

        if($this->debug) {
            print ("<br>Debug retrieveCollection: $classname | query total: $sql_tot $filter <br>");
            print ("<br>Debug retrieveCollection COUNT: $sql_tot sobre campo $campo.<br>");
        }

        $stmt->execute();
        $row = $stmt->fetch();
        $total = $row[$campo];
        $stmt = NULL;


        $sql = "$select $filter $order LIMIT ". ($page-1)*$objPerPage.','.$objPerPage;

        if($this->debug) {
            print ("<br>Debug retrieveCollection: clase: $classname | SQL: $sql <br>");
        }

        $stmt = $this->pdo->query($sql);



// Nota: el primer aproach fue llamar a retrieve() o retrieveColection() para
// hidratar atributos Objeto o Collection, pero esto hace que se ejecute un SELECT
// por cada registro que matchea, lo que es muy costoso en recursos...
// Importante!!! En vez de hidratar así, puedo ejecutar la query e hidratar a todos los atributos
// juntos en un bucle.
// Cómo no lo pensé antes!!!

        // extrae del mapeo el OID
        $oid = $this->classMapper($classname, 'ObjectId');
        // el getter correspondiente
        $getter = $this->classMapper($classname, "Attributes/$oid/getter");

        // Hidrata la colección
        while( $obj = $stmt->fetch(PDO::FETCH_OBJ) ) {
            // crea un Objeto temporal para tener solo
            if($this->debug) {
                    print("<br>Hidratando para Collection $classname con OID = {$obj->$oid} <br>");
                    var_dump($obj);
            }
            // Muy elegante, pero extremadamente costoso :( [de 0.79 a 0.71 segs de diferencia... no fue para tanto... buahhh!!!!]
            //$newobj = $this->retrieve($classname,$obj->$oid);
            $newobj = new $classname;
            // Los mapeos están ahora en un método privado
            $newobj = $this->hydrate($newobj, $obj);

            // Carga el Objeto a la Colección
            $collection->add($newobj);
            $newobj = NULL;
        }

        // Si la Colección a hidratar tiene definidas consultas propias, las busca y ejecuta
        $collatt = $this->getAttibutes($collection);
        foreach ($collatt as $name) {
            // Devuelve el setter mapeado, por default set + atributo
            $setter = $this->classMapper($classname, "Collection/$name/setter");
            $query  = $this->classMapper($classname, "Collection/$name/query");
// Nota: se podrían agregar parámetros a la query para filtros, o intentar usar el mismo filtro seteado
// ver pasaje de parámetros a la query de PDOStatement->execute
            if ( $query != '') {
                $stmt = $this->pdo->prepare($query);
                $stmt->execute();
                $row = $stmt->fetch();
                $collection->$setter($row[$name]);
                $stmt = NULL;
                if($this->debug) {
                    print ("Debug retrieveCollection Totalizadores: $query <br>");
                }
            }
            // Debug	print("<br>AttCol: $setter | $query| $name | R: {$row[$name]}<br>");	print_r($row);
        }

        return $total;
    }

// --------------------------- Métodos Privados ------------------------------

    // inserta un registro con los atributos del Objeto mapeado
    private function insert($clase, $val) {
            // A partir de la clase, obtiene la tabla mapeada.
            $tabla = $this->classMapper($clase, 'Table/Name');
// Debug var_dump($this->tablas); print("Tabla: $tabla <br>");
            // A partir de la clase, obtiene el campo clave
            $pkey  = $this->classMapper($clase, 'Table/PrimaryKey');
            // Si el campo tiene la forma T.campo, se queda solo con el nombre del campo
            if(strpos($pkey, '.') !== false ) {
                    $pkey = substr($pkey, strpos($pkey, '.')+1, strlen($pkey));
            }
            // Valida que la tabla hallada exista o sale con excepción
            if( ! in_array($tabla, $this->tablas) ) {
                     throw new Exception( "Error: Mapeo inexistente en el repositorio. Clase $clase");
            }

//var_dump('ORM',$val);

            $sql = "INSERT INTO `$tabla` SET ";
            foreach ($val as $k => $v) {
                    if ($v !== NULL) {
                        // Si el campo tiene la forma T.campo, se queda solo con el nombre del campo
                            if(strpos($k, '.') !== false ) {
                                    $k = substr($k, strpos($k, '.')+1, strlen($k));
                            }
                            // Si usara BindParam en PDO, no habría problemas con las comillas...
                            //$sql.= " `$k`='$v',";
                            /*$v = addslashes($v);
                            $sql.= " `$k`='$v',";*/
                            $sql.= " `$k`=\"$v\",";
                    } else {
                        // Soporte para generar INSERT con los campos NULL
                        // Importante: salvo para los OID, porque si no, falla la query en los autoincrementales
                        if($k != $pkey) {
                            $sql.= " `$k`=NULL,";
                        }
                    }
            }
            // Borra la última coma
            $sql = substr($sql,0,strlen($sql)-1);

            if($this->debug) {
                    print("Debug insert: $sql <br>");
            }
            // Devuelve el último ID generado...
            if( $this->pdo->exec($sql) === false ) {
                    $error = $this->pdo->errorInfo();
                    throw new Exception( "Error: class $clase | {$error[0]} - {$error[1]} : {$error[2]}." );
            }
            // Devuelve el valor del OID
            return $this->pdo->lastInsertId();
    }

    // actualiza un registro con los atributos del Objeto mapeado
    private function update($clase, $val, $pk="id") {
            // A partir de la clase, obtiene la tabla mapeada.
            $tabla = $this->classMapper($clase, 'Table/Name');
            // Valida que la tabla hallada exista o sale con excepción
            if( ! in_array($tabla, $this->tablas) ) {
                     throw new Exception( "Error: Mapeo inexistente en el repositorio. Clase $clase");
            }
            // A partir de la clase, obtiene el campo clave
            $pkey  = $this->classMapper($clase, 'Table/PrimaryKey');
            // Si el campo tiene la forma T.campo, se queda solo con el nombre del campo
            if(strpos($pkey, '.') !== false ) {
                    $pkey = substr($pkey, strpos($pkey, '.')+1, strlen($pkey));
            }


            $sql = "UPDATE `$tabla` SET ";
            // Los $K pueden ser mapeados
            foreach ($val as $k => $v) {
                    if ($this->classMapper($clase, "Attributes/$k/field") != '') {
                            $campo = $this->classMapper($clase, "Attributes/$k/field");
                    } else {
                            $campo = $k;
                    }
                    if ($v != NULL) {
    // Debug
    //print("Debug update: clase = $clase | tabla = $tabla | campo = $campo | k = $k | Mapper = {$this->classMapper($classname, "Attributes/$k/field")}<br>");
                            // Arma la query
                            // Si usara BindParam en PDO, no habría problemas con las comillas...
                            /*$v = addslashes($v);
                            $sql.= " `$campo`='$v',";*/
                            $sql.= " `$campo`=\"$v\",";
                    } else {
                            $sql.= " `$campo`=NULL,";
                    }
            } // foreach
            // Borra la última coma
            $sql = trim($sql);
            $sql = substr($sql,0,strlen($sql)-1);
            $sql .= " WHERE `$pkey`='$val[$pk]' ";
            if($this->debug) {
                    print("Debug update: $sql <br>");
            }
            //return $this->pdo->exec($sql);
            if( $this->pdo->exec($sql) === false ) {
                    $error = $this->pdo->errorInfo();
                    throw new Exception( "Error Update: class $clase | {$error[0]} - {$error[1]} : {$error[2]}." );
            }
            // Devuelve el valor del OID
            return $val[$pk];
    }

    // Arma el SELECT correspondiente al Mapeo de la clase
    private function select($classname, $valoid='') {
            // A partir de la clase, obtiene la tabla mapeada.
            $tabla = $this->classMapper($classname, 'Table/Name');
            // Puede tener mapeado un SELECT distinto al default
            $select = $this->classMapper($classname, 'Table/Select');
            $where = ' WHERE ';
            // Si el select está mapeado, puede tener ya un WHERE... en ese caso hay que poner un AND en lugar de un WHERE
            if($select == '') { $select = "SELECT * FROM `$tabla`"; }
            else {
                    if( strrpos($select, 'WHERE') !== false ) {
                            $where = ' AND ';
                    }
            }
            if($valoid != '') {
                    // A partir de la clase, obtiene el campo clave
                    $pkey  = $this->classMapper($classname, 'Table/PrimaryKey');
                    $valoid = "$where $pkey='$valoid'";
            }
            //return "$select $where $pkey=$id ";
            return "$select $valoid ";
    }

    private function hydrate(&$obj, $queryObj) {
        // Hidrata al Objeto $obj con los datos del STDObj devuelto por una query PDO

        // Obtiene la clase del Objeto para hallar los mapeos
        $classname = get_class($obj);

// Atención! Nueva forma de a[]
// setter => valor

        // Vector con los atributos
        $att = $this->getAttibutes($classname);
        foreach($att as $name) {
                            $setter = $this->classMapper($classname, "Attributes/$name/setter");
            $campo = $this->classMapper($classname, "Attributes/$name/field");

            // Debugprint("<br />Setter: $setter | Campo: $campo<br />");var_dump($queryObj);

// <Relacion>
            // Si está mapeada una relación y se piden Objetos anidados, hidrata al objeto relacionado
            $type = '';
            $relation  = $this->classMapper($classname, "Attributes/$name/relation", $type);
            $att_name = '';
            $reference  = $this->classMapper($classname, "Attributes/$name/reference", $att_name);

            // Si no se piden objetos anidados, fuerza a no entrar nunca al if (pero sí tiene que entrar al else
            if (! $this->recursiveObject ) {
                $relation = '';
                $reference = '';
            }

            if ( ($relation != '') && ($reference != '' ) ) {

                if($this->debug) {
                    print ("Debug Hydrate retrieve relation: Relacion: $relation - Clase: $reference -
                                    Name: $name - Campo: $campo - Setter: $setter - Obj: ".var_dump($queryObj)." <br>");
                }

                // Hidratación
                //$objrel = $this->retrieve("$reference",$obj->$name);
                switch($relation) {
                    case '1-N': // hidrata la colección relacionada.
                        // Guarda temporalmente el filtro actual (ahora es un vector)
                        $tmpfa = $this->filterAttrib;
                        $tmpfv = $this->filterValue;
                        $tmpfe = $this->filterExac;
                        // Guarda temporalmente el orden actual
                        $tmpco = $this->collectionOrder;
                        $tmpca = $this->collectionAsc;
                        // Halla el ID del Objeto de referencia
                        $oid = $this->classMapper($classname, 'ObjectId');
                        $id = $queryObj->$oid;
                        // Aplica el filtro de los objetos relacionados
                        $this->resetCollectionFilter(); // pone a nada los filtros previamente
                        $this->setCollectionFilter($att_name, $id);
                        // Remueve filtros existentes
                        $this->setCollectionOrder('');
                        $getter1 = $this->classMapper($classname, "Attributes/$name/getter");

                        if($this->debug) {print("Hydrate 1-N: Ref: $reference | Att.Name: $att_name | ID 1: $id} | Getter: $getter1 <br>"); }

                        // Only variables should be passed by reference ...
                        //$this->retrieveCollection($reference, $obj->{$getter1}(), 1, PersistentMaxObjectPerPage);
                        $referenceCol = $obj->{$getter1}();
                        $this->retrieveCollection($reference, $referenceCol, 1, PersistentMaxObjectPerPage);

                        // Restaura los filtros existentes (los filtros ahora son vectores
                        //$this->setCollectionFilter($tmpfa, $tmpfv,$tmpfe);
                        $this->filterAttrib = $tmpfa;
                        $this->filterValue = $tmpfv;
                        $this->filterExac = $tmpfe;
                        // Restaura el orden existente
                        $this->setCollectionOrder($tmpco, $tmpca);
                        // Por ser una colección, no agrega una entrada a $a
                        break;
                    default: // hidrata el objeto relacionado
                        //var_dump($queryObj->$campo);
                        $objrel = $this->retrieve($reference, $queryObj->$campo);
//print_r($objrel);
                        //$a[$campo] = $objrel;
                        $a[$setter] = $objrel;
                } // switch($relation)
            } else {
                // Protección contra campos ausentes (por relaciones 1-N)
                if ($this->recursiveObject ) {
                    $a[$setter] = $queryObj->$campo;
                } else {
                    if(isset($queryObj->$campo)) {
                        $a[$setter] = $queryObj->$campo;
                    }
                }
            }
/*			// Descubre si el atributo es un Objeto
            if (is_object($name)) { print("$name es Objeto<br>"); }
            else {print("$name NO es Objeto<br>"); } */

// </Relacion>

        } // foreach($att as $name)
// Debug print("<pre>");print_r($a);print("</pre>");
//Debug var_dump($obj);var_dump($a);

        if($this->debug) {
            if(isset($id)) {
                   print ("Debug Hydrate retrieve: hidratando (new) $classname - OID $id<br>");
                } else {
                   print ("Debug Hydrate retrieve: hidratando (new) $classname <br>");
                }
        }

//		$newobj = new $classname;
        foreach ($a as $k => $v) {
            // Debug
//print(" <br> <b>$k</b> | $v | {$queryObj->$k} # ->$v() par: $k<br>");
            //$newobj->$v($queryObj->$k);
            $obj->$k($v);
        }

        return $obj;
    }

    // ----------------------- Reflection ------------------------------

    // Devuelve un vector con los atributos hallados de la Clase
    private function getAttibutes($class) {
	      $res = array();
	      try {
	              $ref = new ReflectionClass($class);
	      } catch (Exception $e) {
	              throw new Exception($e->getMessage()." | $class");
	      }
	      $temp = $ref->getProperties();
	      foreach ($temp as $rp) {
	              // Obtiene Atributos
	              $res[] = $rp->getName();
	      }
	      $ref = NULL;
	      return $res;
    }

    // -------------------- mapeo --------------------
    // Este mapeo bien podría sacarse a otra clase con la que Persistent colabore.
    private function classMapper($clase, $map, &$tipo=NULL){
            // Devuelve el mapeo de la clase requerido. Si no lo encuentra, devuelve defaults

            // Si no hay archivo de configuracion
            if ( ! file_exists($this->configxml)) {
                    return $clase;
            } else {

                    /* * Pasado al constructor...
                     * tenerlo acá, implica que para cada mapeo se tiene
                     * que volver a leer del disco el XML, con lo que el
                     * tiempo de ejecución en mapeos complejos e multiplica por ocho (7)
                     $doc = new DOMDocument;
                    // We don't want to bother with white spaces
                    $doc->preserveWhiteSpace = false;
                    $doc->Load($this->configxml); */

                    // Debug	print("classMapper: $clase");
                    $xpath = new DOMXPath($this->xmldoc);
                    // Arma la query XPATH
                    $query = "//Persistent/$clase/$map";
						  // Los ( y ) rompen la query; sacarlos permite salvajadas en los filtros ;-)
						  $query = str_replace(array('(',')',' ',"\n",','), '', $query);

// Debug print("Debug XQuery: $query <br>");
                    // ejecuta la query XPATH
                    $entries = $xpath->query($query);
// Debug var_dump($entries);
							if(! $entries) {
								$entries = NULL;
								$entry = NULL;
							} else {
	                    // La respuesta es una sola cosa
   	                 $entry = $entries->item(0);
                    }
                    // Obtiene la respuesta
                    $valor = NULL;
                    if(isset ($entry->nodeValue)) {
                        $valor = $entry->nodeValue;
                    }
                    // Obtiene el atributo "type"
                    if ($entry != NULL) { // (se pregunta por muchos mapeos inexistentes: los defaults)
                            // puede ser "type" o "attibrute_name"
                            $tipo = $entry->getAttribute('type');
                            $tipo .= $entry->getAttribute('attribute_name');
                    }
                    // Si la consulta es por Attributes/$name/reference y está vacía, devuelve el nombre de la clase
                    $refatttmp = explode('/', $map);
                    if(isset ($refatttmp[2]) && ($refatttmp[2] == 'reference')&&($tipo=='')) {
                            $tipo = $clase;
                    }
                    // El mapeo del count tiene atributo fieldname
/*			if(($map == 'Table/Count')&&($entry != NULL)) {
                            $tipo = $entry->getAttribute('fieldname');
                            if($tipo == '') { $tipo = 'objtot'; }
                    }*/
                    if($map == 'Table/Count') {
                            $tipo = 'objtot';
                            if($entry != NULL) {
                                    $tmptipo = $entry->getAttribute('fieldname');
                                    if($tmptipo != '') { $tipo = $tmptipo; }
                            }
                    }

                    // Libera recursos... o al menos lo intenta
                    unset($entry);
                    unset($xpath);
                    //unset($doc);

                    if (($valor != NULL)||$valor != '') {
                            return $valor;
                    } else {
                            // Si no se encuentra qué devolver, debe devolver el default
                            switch($map) {
                                    case 'Table/Name':
                                                            return $clase;
                                                            break;
                                    case 'Table/PrimaryKey':
                                                            return 'id';
                                                            break;
                                    case 'Table/Count':
                                                            $pkey  = $this->classMapper($clase, 'Table/PrimaryKey');
                                                            $tabla = $this->classMapper($clase, 'Table/Name');
                                                            $sql_count = "SELECT COUNT($pkey) as $tipo FROM `$tabla`";
                                                            return $sql_count;
                                                            break;
                                    case 'ObjectId':
                                                            return 'id';
                                                            break;
                                    default: // Los atributos
                                            $atttmp = explode("/", $map);
                                            if(isset ($atttmp[2])) {
                                                switch($atttmp[2]) {
                                                    case 'setter':
                                                                    return 'set'.$atttmp[1];
                                                                    break;
                                                    case 'getter':
                                                                    return 'get'.$atttmp[1];
                                                                    break;
                                                    case 'field':
                                                                    return $atttmp[1];
                                                                    break;
                                                }
                                            }
                            } // switch($map)
                    } // else if (! $valor == NULL) {
            } // if ( file_exists($this->configxml)) {
    } // function classMapper(

} // de la clase
